import itertools
import os
import importlib
from pathlib import Path

from ..utils import Utils

config = None
UTILITIES = None


def import_utilities():
    global UTILITIES
    from . import __all__

    UTILITIES = [importlib.import_module(f".{utility}", package=__package__) for utility in __all__]


def file_setup(utility):
    file_location = config.input_folder_name

    if config.output_folder_name != 'DEFAULT':
        output_folder = config.output_folder_name
    else:
        output_folder = utility.FOLDER_NAME

    if config.base_folder in ("", " "):
        print('Select the directory which contains the subject folders.')
        base_sub_location = Utils.file_browser(title='Select the directory which contains the subject folders')

    else:
        base_sub_location = config.base_folder

    # Create dictionary from each participant directory
    participant_dir, _ = Utils.find_participant_dirs(base_sub_location)
    participants = {participant: None for participant in participant_dir}

    for participant in participants:
        # Find all nifti and analyze files
        participants[participant] = Utils.find_files(f"{participant}/{file_location}", "hdr", "nii.gz", "nii")

        # Make directory to save results
        if isinstance(output_folder, list):
            for folder in output_folder:
                Utils.check_and_make_dir(f"{participant}/{folder}", delete_old=True)
        else:
            Utils.check_and_make_dir(f"{participant}/{output_folder}", delete_old=True)

    return participants, output_folder, file_location


def utility_handler(participants, output_folder, file_location, utility):
    if config.multicore_processing:
        pool = Utils.start_processing_pool()
    else:
        pool = None

    if config.verbose and isinstance(output_folder, list):
        direcs = ', '.join(output_folder)
        print(f'\nSaving output in directories: {direcs}')
    elif config.verbose and len(output_folder) == 1:
        print(f'\nSaving output in directory: {output_folder}')

    for participant_dir, files in participants.items():
        if config.verbose:
            print(f"\nRunning '{utility.UTILITY_NAME}' utility on participant: {participant_dir.split('/')[-1]}"
                  f"\n      Running utility on {len(files)} files")

        iterable = zip(files,
                       itertools.repeat(participant_dir),
                       itertools.repeat(file_location),
                       itertools.repeat(output_folder),
                       itertools.repeat(utility),
                       itertools.repeat(config))

        if config.multicore_processing:
            return_val = list(pool.starmap(run_utility, iterable))

        else:
            return_val = list(itertools.starmap(run_utility, iterable))

        if return_val[0] is not None:
                iterable = zip(files,
                               itertools.repeat(participant_dir),
                               itertools.repeat(file_location),
                               itertools.repeat(output_folder),
                               itertools.repeat(utility),
                               itertools.repeat(config),
                               itertools.repeat(return_val))

                if config.multicore_processing:
                    list(pool.starmap(run_utility, iterable))

                else:
                    list(itertools.starmap(run_utility, iterable))

    if config.multicore_processing:
        Utils.join_processing_pool(pool, restart=False)


def run_utility(file, participant_dir, file_location, output_folder, utility, config, *args):
    no_ext_file = Utils.strip_ext(file)
    file = f"{participant_dir}/{file_location}/{file}"
    base_sub_location = Path(participant_dir).parents[0]
    participant_name = participant_dir.split('/')[-1]

    if config.verbose and not args:
        print(f'        Running utility on file: {no_ext_file}')

    return_val = utility.run(config, file, no_ext_file, base_sub_location, participant_name, participant_dir, output_folder, return_val=args)

    return return_val


def find_current_utility(func):
    for utility in UTILITIES:
        if func == utility.UTILITY_NAME:
            return utility


def HOUSE(cfg, func):
    global config
    config = cfg

    import_utilities()
    utility = find_current_utility(func)
    participants, output_folder, file_location = file_setup(utility)
    utility_handler(participants, output_folder, file_location, utility)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fRAT', 'fRAT.HOUSE', 'fRAT.utils']

package_data = \
{'': ['*'],
 'fRAT': ['configuration_profiles/*',
          'configuration_profiles/maps/*',
          'configuration_profiles/roi_analysis/*',
          'images/*']}

install_requires = \
['altgraph==0.17.2',
 'bootstrapped==0.0.2',
 'brotli==1.0.9',
 'certifi==2022.12.7',
 'chardet==3.0.4',
 'charset-normalizer==2.0.12',
 'ci-info==0.2.0',
 'click==7.1.1',
 'contourpy==1.0.7',
 'cycler==0.10.0',
 'cython==0.29.21',
 'dash-core-components==1.15.0',
 'dash-html-components==1.1.2',
 'dash-renderer==1.9.0',
 'dash-table==4.11.2',
 'dash==1.19.0',
 'dataclasses==0.6',
 'decorator==4.4.2',
 'descartes==1.1.0',
 'dill==0.3.6',
 'dominate==2.6.0',
 'etelemetry==0.2.2',
 'filelock==3.0.12',
 'flask-compress==1.8.0',
 'flask==1.1.2',
 'fonttools==4.38.0',
 'fslpy==3.4.0',
 'funcsigs==1.0.2',
 'future==0.18.2',
 'gevent==22.10.2',
 'greenlet==2.0.2',
 'h5py==3.8.0',
 'idna==2.9',
 'imageio==2.9.0',
 'isodate==0.6.0',
 'itsdangerous==1.1.0',
 'jinja2==2.11.3',
 'joblib==1.2.0',
 'kiwisolver==1.4.4',
 'looseversion==1.0.1',
 'lxml==4.9.2',
 'macholib==1.15.2',
 'markupsafe==1.1.1',
 'matplotlib==3.6.3',
 'mizani==0.8.1',
 'multiprocess==0.70.14',
 'networkx==2.4',
 'neurdflib==5.0.1',
 'nibabel==3.2.1',
 'nilearn==0.9.0',
 'nipype==1.8.4',
 'numpy==1.22.3',
 'packaging==23.0',
 'palettable==3.3.0',
 'pandas==1.5.3',
 'pathos==0.2.6',
 'patsy==0.5.3',
 'pillow==9.3.0',
 'pip-autoremove==0.10.0',
 'plotly==4.14.3',
 'plotnine==0.10.1',
 'pox==0.2.8',
 'ppft==1.6.6.2',
 'prov==1.5.3',
 'pydot==1.4.1',
 'pydotplus==2.0.2',
 'pyinstaller-hooks-contrib==2022.2',
 'pyinstaller==4.9',
 'pyparsing==2.4.6',
 'python-dateutil==2.8.1',
 'pytz==2022.7.1',
 'pywavelets==1.4.1',
 'rdflib==5.0.0',
 'requests==2.27.1',
 'retrying==1.3.3',
 'scikit-image==0.19.3',
 'scikit-learn==1.2.1',
 'scipy==1.10.0',
 'simplejson==3.17.0',
 'six==1.14.0',
 'sklearn==0.0',
 'statsmodels==0.13.5',
 'tabulate==0.8.9',
 'tenacity==8.2.0',
 'threadpoolctl==2.1.0',
 'tifffile==2021.7.30',
 'toml==0.10.2',
 'traits==6.3.2',
 'urllib3==1.26.5',
 'werkzeug==1.0.1',
 'xlrd==1.2.0',
 'xmltodict==0.12.0',
 'zope-event==4.5.0',
 'zope-interface==5.2.0']

entry_points = \
{'console_scripts': ['fRAT = fRAT.__main__:start_gui']}

setup_kwargs = {
    'name': 'frat-brain',
    'version': '1.6.3',
    'description': 'Application for ROI fMRI data analysis.',
    'long_description': '<img src="https://github.com/elliohow/fMRI_ROI_Analysis_Tool/blob/master/docs/images/fRAT.gif?raw=true" width=500>\n\n# fRAT - fMRI ROI Analysis Tool\n[![status](https://joss.theoj.org/papers/cc9c0cb3b12abaf30c8381728d3229d7/status.svg)](https://joss.theoj.org/papers/cc9c0cb3b12abaf30c8381728d3229d7)\n[![PRs Welcome](https://img.shields.io/badge/PRs-welcome-brightgreen.svg?style=flat-square)](http://makeapullrequest.com) \n[![GitHub license](https://img.shields.io/hexpm/l/plug?style=flat-square)](https://github.com/elliohow/fMRI_ROI_Analysis_Tool/blob/master/LICENSE)\n[![Github release (latest by date)](https://img.shields.io/github/v/release/elliohow/fmri_roi_analysis_tool?style=flat-square)](https://github.com/elliohow/fmri_roi_analysis_tool/releases/latest)\n[![Github issues](https://img.shields.io/github/issues/elliohow/fmri_roi_analysis_tool?style=flat-square)](https://github.com/elliohow/fmri_roi_analysis_tool/issues)\n[![Documentation](https://img.shields.io/readthedocs/fmri-roi-analysis-tool)](https://fmri-roi-analysis-tool.readthedocs.io/en/latest/)\n\nfRAT is an open-source python-based GUI application used to simplify the processing and analysis of fMRI data by\nconverting voxelwise maps into ROI-wise maps. An installation of FSL is required in order to use fRAT.\n\n> fRAT is written using **Python** for **MacOS, Linux and WSL2**.\n\nDocumentation:\n\n[Home page](https://fmri-roi-analysis-tool.readthedocs.io)\n\n[Installation instructions](https://fmri-roi-analysis-tool.readthedocs.io/en/latest/installation.html)\n\n[ROI analysis tutorial](https://fmri-roi-analysis-tool.readthedocs.io/en/latest/tutorials/Basic-ROI-analysis.html)\n\n## Citation\n\n**When using fRAT, please include the following citation:**\n\nHowley, E., Francis, S., & Schluppeck, D. (2023). fRAT: an interactive, Python-based tool for region-of-interest summaries of functional imaging data. Journal of Open Source Software, 8(85), 5200. https://doi.org/10.21105/joss.05200\n\n## Reporting bugs\n\nTo report a bug or suggest a new feature, please go to [fRAT\'s Issues](https://github.com/elliohow/fMRI_ROI_Analysis_Tool/issues/new/choose).\n\nFor other questions, issues or discussion please go to [fRAT\'s Discussions](https://github.com/elliohow/fMRI_ROI_Analysis_Tool/discussions).\n\n## Contributing to the project\n\nIf you\'d like to contribute to the project please read our [contributing guidelines](https://github.com/elliohow/fMRI_ROI_Analysis_Tool/blob/master/CONTRIBUTING.md). Please also read through our [code of conduct](https://github.com/elliohow/fMRI_ROI_Analysis_Tool/blob/master/CODE_OF_CONDUCT.md).\n\n## Versioning\nWe use [Semantic versioning](http://semver.org/) for versioning. For the versions available, see the\n[tag list](https://github.com/elliohow/fMRI_ROI_Analysis_Tool/tags) for this project.\n\n## Licensing\nThis project uses the Apache 2.0 license. For the text version of the license see\n[here](https://github.com/elliohow/fMRI_ROI_Analysis_Tool/blob/master/LICENSE). \nPrior to version 1.0.0, this project used an MIT license.\n\n## Images\n<img src="https://github.com/elliohow/fMRI_ROI_Analysis_Tool/blob/master/docs/images/GUI.png?raw=true" title="Example of the fRAT GUI" width=700>\n\n<img src="https://github.com/elliohow/fMRI_ROI_Analysis_Tool/blob/master/docs/images/HTML_report.png?raw=true" title="Example of a HTML report output by fRAT" width=600>\n',
    'author': 'Elliot Howley',
    'author_email': 'elliohow@hotmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://fmri-roi-analysis-tool.readthedocs.io/en/latest/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



import datetime
import itertools
import shutil
import sys
import time
from copy import deepcopy
from pathlib import Path

from fRAT.utils import *
from fRAT._version import __version__


def fRAT(config_filename, path=None):
    start_time = time.perf_counter()
    Utils.checkversion(__version__)

    config_path = f'{Path(os.path.abspath(__file__)).parents[0]}/configuration_profiles/roi_analysis/'
    config, orig_path = load_config(config_filename, path=path)
    config = argparser(config)

    # CompareOutputs.run(config)  # TODO THIS IS TEST CODE
    # sys.exit()

    if config.verbose and config.run_analysis and config.run_plotting and config.run_statistics:
        print(f"\n--- Running analysis, plotting and statistics steps ---")

    # Run the ROI analysis
    if config.run_analysis:
        analysis(config, config_path, config_filename)

    # Plot the results
    if config.run_plotting:
        plotting(config, config_path, config_filename, orig_path)

    # Run the statistical analysis
    if config.run_statistics:
        if config.verbose:
            print('\n------------------\n--- Statistics ---\n------------------')

        Utils.chdir_to_output_directory('Statistics', config)
        statistics_main(config, config_path, config_filename)

    os.chdir(orig_path)  # Reset path

    if config.verbose:
        elapsed = datetime.timedelta(seconds=time.perf_counter() - start_time)
        elapsed_min = elapsed.seconds // 60
        elapsed_sec = elapsed.seconds % 60
        print(f"--- Completed in {elapsed_min} minutes {elapsed_sec} seconds ---\n\n")


def load_config(config_filename, path=None):
    orig_path = Path(os.path.abspath(__file__)).parents[0]

    # Reload config file incase GUI has changed it
    config = Utils.load_config(f'{Path(os.path.abspath(__file__)).parents[0]}/configuration_profiles/roi_analysis',
                               config_filename, path=path)
    config_check(config)

    return config, orig_path


def argparser(config):
    # Check arguments passed over command line
    args = Utils.argparser()

    if args.make_table == 'true':
        from __main__ import make_table
        make_table()
        sys.exit()
    else:
        args.__dict__.pop('make_table')

    for arg in args.__dict__:
        if args.__dict__[arg] is not None:
            user_arg = Utils.convert_toml_input_to_python_object(args.__dict__[arg])
            config.__dict__[arg] = user_arg

    config = Utils.clean_config_options(config)

    return config


def plotting(config, config_path, config_filename, orig_path):
    if config.verbose:
        print('\n----------------\n--- Plotting ---\n----------------')

    # Plotting
    Utils.chdir_to_output_directory('Plotting', config)
    Figures.make_figures(config, config_path, config_filename)

    # Create html report
    create_html_report(str(orig_path))
    if config.verbose:
        print('Created html report.')


def analysis(config, config_path, config_filename):
    if config.verbose:
        print('\n----------------\n--- Analysis ---\n----------------')

    if config.multicore_processing:
        pool = Utils.start_processing_pool()
    else:
        pool = None

    # Run class setup
    participant_list, matched_brains = Environment_Setup.setup_analysis(config, config_path, config_filename, pool)

    Utils.move_file(config.parameter_file, os.getcwd(), os.getcwd() + f"/{Environment_Setup.save_location}",
                    copy=True, parameter_file=True)

    if config.verbose:
        print('\n--- Running individual analysis ---')

    brain_list = []
    for participant in participant_list:
        brain_list.extend(participant.run_analysis(pool))

        if config.file_cleanup == 'move' and not participant.all_files_ignored:
            shutil.move(f"{participant.save_location}/motion_correction_files",
                        f"{participant.save_location}Intermediate_files/motion_correction_files")

        elif config.file_cleanup == 'delete' and not participant.all_files_ignored:
            shutil.rmtree(f"{participant.save_location}/motion_correction_files")

    matched_brains = run_pooled_analysis(brain_list, matched_brains, config, pool)

    calculate_cost_function_and_displacement_values(participant_list, brain_list, config, pool)
    atlas_scale(matched_brains, config, pool)

    if config.multicore_processing:
        Utils.join_processing_pool(pool, restart=False)


def run_pooled_analysis(brain_list, matched_brains, config, pool):
    if config.verbose:
        print('\n--- Running pooled analysis ---')

    for parameter_comb in matched_brains:
        for brain in brain_list:
            try:
                if brain.no_ext_brain in parameter_comb.brains[brain.participant_name]:
                    parameter_comb.ungrouped_summarised_results.append(brain.roi_results)
                    parameter_comb.ungrouped_raw_results.append(brain.raw_results_loc)
                    parameter_comb.participant_grouped_summarised_results[brain.participant_name].append(brain.roi_results)

            except KeyError:
                pass

    # Save each raw and overall results for each parameter combination
    iterable = zip(matched_brains,
                   itertools.repeat("compile_results"),
                   itertools.repeat(os.getcwd()),
                   itertools.repeat(config))

    if config.multicore_processing:
        matched_brains = pool.starmap(Utils.instance_method_handler, iterable)
    else:
        matched_brains = list(itertools.starmap(Utils.instance_method_handler, iterable))

    # Compile the overall results for every parameter combination
    construct_combined_results(MatchedBrain.save_location, subfolder='session averaged')
    construct_combined_results(MatchedBrain.save_location, subfolder='participant averaged')

    return matched_brains


def calculate_cost_function_and_displacement_values(participant_list, brain_list, config, pool):
    if config.verbose:
        print('\n--- Calculating cost function and displacement values ---')

    vals = []

    for counter, participant in enumerate(participant_list):
        if config.verbose:
            print(f'Calculating cost function value for anatomical file: {participant.anat_brain_no_ext}')

        if participant.all_files_ignored:
            continue

        df = participant.calculate_anat_flirt_cost_function()
        vals.append(df)

    # Set arguments to pass to fmri_get_additional_info
    iterable = zip(brain_list, itertools.repeat("fmri_get_additional_info"),
                   range(len(brain_list)),
                   itertools.repeat(len(brain_list)),
                   itertools.repeat(config))

    if config.multicore_processing:
        results = pool.starmap(Utils.instance_method_handler, iterable)
    else:
        results = list(itertools.starmap(Utils.instance_method_handler, iterable))

    vals.extend(results)

    if config.verbose:
        print(f'\nSaving dataframe as additional_info.csv')

    df = pd.concat(vals).replace([np.nan], [None]).sort_values(['Participant', 'File']).reset_index(drop=True)

    with open(f"{Environment_Setup.save_location}additional_info.csv", 'w') as file:
        df.to_csv(path_or_buf=file)


def run_atlas_scale_function_for_each_statistic(first_combination, matched_brains, statistic_number,
                                                pool, config, data):
    if data == 'Session averaged':
        roi_stats = deepcopy(first_combination.session_averaged_results[statistic_number, :])

    else:
        roi_stats = deepcopy(first_combination.participant_averaged_results[statistic_number, :])

    for parameter_combination in matched_brains:
        roi_stats = find_highest_value_from_overall_results(parameter_combination, roi_stats, statistic_number,
                                                            data=data)

    # Set arguments to pass to atlas_scale function
    iterable = zip(matched_brains, itertools.repeat("atlas_scale"), itertools.repeat(roi_stats),
                   range(len(matched_brains)), itertools.repeat(len(matched_brains)),
                   itertools.repeat(statistic_number), itertools.repeat(Environment_Setup.atlas_path),
                   itertools.repeat(data), itertools.repeat(config))

    # Run atlas_scale function and pass in max roi stats for between brain scaling
    if config.multicore_processing:
        pool.starmap(Utils.instance_method_handler, iterable)

    else:
        list(itertools.starmap(Utils.instance_method_handler, iterable))


def find_highest_value_from_overall_results(parameter_combination, roi_stats, statistic_number, data):
    if data == 'Session averaged':
        results = parameter_combination.session_averaged_results[statistic_number, :]
    else:
        results = parameter_combination.participant_averaged_results[statistic_number, :]

    for counter, roi_stat in enumerate(results):
        if roi_stat > roi_stats[counter]:
            roi_stats[counter] = roi_stat

    return roi_stats


def atlas_scale(matched_brains, config, pool):
    """Save a copy of each statistic for each ROI from the first brain. Then using sequential comparison
       find the largest statistic values for each ROI out of all the brains analyzed."""
    if config.verbose:
        print('\n--- Atlas scaling ---')

    # Make directory to store scaled brains
    Utils.check_and_make_dir(f"{os.getcwd()}/{MatchedBrain.save_location}NIFTI_ROI/")
    Utils.check_and_make_dir(f"{os.getcwd()}/{MatchedBrain.save_location}NIFTI_ROI/Session_averaged_results/")
    Utils.check_and_make_dir(f"{os.getcwd()}/{MatchedBrain.save_location}NIFTI_ROI/Participant_averaged_results/")

    first_combination = next(iter(matched_brains))

    if config.verbose:
        print('Creating NIFTI ROI images using session averaged data.\n')

    for statistic_number in range(len(first_combination.session_averaged_results)):
        run_atlas_scale_function_for_each_statistic(first_combination, matched_brains, statistic_number,
                                                    pool, config, data='Session averaged')
    if config.verbose:
        print('\nCreating NIFTI ROI images using participant data.\n')

    for statistic_number in range(len(first_combination.participant_averaged_results)):
        run_atlas_scale_function_for_each_statistic(first_combination, matched_brains, statistic_number,
                                                    pool, config, data='Participant averaged')


def config_check(config):
    pass


if __name__ == '__main__':
    fRAT()
